"""This module exports the renderer class for board objects and their variants.
"""

from collections import defaultdict
from collections.abc import Sequence

from rich import box
from rich.console import Group
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from .task_renderer import TaskRenderer
from ..config import STATUS_COLOURS, STATUS_NAMES
from ..enums import Status
from ..models import Board, Task


class BoardRenderer:
    """Class responsible for defining how a board should be displayed.
    """

    @staticmethod
    def _create_base_table(title: str, *, board_column: bool = False) -> Table:
        """Generate an empty rich table object.

        :param title: table name
        :param board_column: whether to include board column or not
        :return: table object
        """
        table = Table(title=title, box=box.DOUBLE, expand=True,
                      show_lines=True)

        if board_column:
            table.add_column('Board')

        for s in Status:
            table.add_column(
                f'[{STATUS_COLOURS[s]}]{STATUS_NAMES[s]}[/]',
                ratio=1
            )

        return table

    @staticmethod
    def _group_tasks_by_status(tasks: Sequence[Task]) -> dict[Status, list[Task]]:
        """Return the tasks grouped by status.

        :param tasks: list of tasks to group
        :return: mapping of status and tasks
        """
        groups = defaultdict(list)

        for task in tasks:
            groups[task.status].append(task)

        return groups

    @classmethod
    def to_kanban(cls, board: Board) -> Table:
        """Return a rich table to display a Kanban board from a board object.

        :param board: board object
        :return: rich table
        """
        statuses = cls._group_tasks_by_status(board.tasks)

        table = cls._create_base_table(f'\\[{board.id}] {board.name}')

        table.add_row(*[
            Group(*(TaskRenderer.to_panel(t) for t in statuses[s]))
            for s in Status
        ])

        return table

    @classmethod
    def to_kanban_swimlanes(cls, boards: Sequence[Board]) -> Table:
        """Return a rich table to display a Kanban board from multiple board
        objects.

        :param boards: list of board objects
        :return: rich table
        """
        table = cls._create_base_table('All active work', board_column=True)

        for board in boards:
            statuses = cls._group_tasks_by_status(board.tasks)

            table.add_row(
                Text(f'\n[{board.id}] {board.name}', style='cyan',
                     no_wrap=True),
                *[
                    Group(*(TaskRenderer.to_panel(t) for t in statuses[s]))
                    for s in Status
                ]
            )

        return table

    @classmethod
    def kanban_from_tasks(cls, title: str, tasks: Sequence['Task']) -> Table:
        """Generate a rich table to display a Kanban board from a list of tasks.

        :param title: table title
        :param tasks: list of tasks to include
        :return: rich table
        """
        table = cls._create_base_table(title)
        statuses = cls._group_tasks_by_status(tasks)

        table.add_row(*[
            Group(*(TaskRenderer.to_panel(t) for t in statuses[s]))
            for s in Status
        ])

        return table



"""Commands responsible for managing boards.
"""

from typing import Annotated

import typer
from sqlalchemy.orm import Session

from ..console import console
from ..db.engine import engine
from ..exceptions import BoardNotFoundError
from ..renderers.board_renderer import BoardRenderer
from ..renderers.message_renderer import MessageRenderer
from ..repos.board_repo import BoardRepository
from ..repos.task_repo import TaskRepository
from ..services.board_service import BoardService


app = typer.Typer(name='board', help='Manage boards.', no_args_is_help=True)


@app.command()
def add(name: Annotated[str, typer.Argument(help='Board name.')]):
    """Create a new board.
    """
    with Session(engine) as session:
        board_repo = BoardRepository(session)
        tasks_repo = TaskRepository(session)
        service = BoardService(board_repo, tasks_repo)

        board = service.create_board(name)
        session.commit()

        console.print(MessageRenderer.success(
            f'Created board "{board.name}".'))


@app.command(help='Rename a board.')
def rename(id: Annotated[int, typer.Argument(help='Board ID.')],
           name: Annotated[str, typer.Argument(help='New name.')]):
    """Rename a board.
    """
    with Session(engine) as session:
        board_repo = BoardRepository(session)
        tasks_repo = TaskRepository(session)
        service = BoardService(board_repo, tasks_repo)

        try:
            board = service.rename_board(id, name)
            session.commit()
        except BoardNotFoundError:
            return console.print(MessageRenderer.error('Board not found.'))

        console.print(
            MessageRenderer.success(f'Renamed board to "{board.name}".'))


@app.command()
def rm(id: Annotated[int, typer.Argument(help='Board ID.')],
       force: Annotated[bool, typer.Option(
           '--force', '-f',
           prompt='Are you sure you want to delete the board?',
           help='Force deletion without confirmation.')] = False):
    """Delete existing board.

    If --force is not used, will ask for confirmation.
    """
    if not force:
        return

    with Session(engine) as session:
        board_repo = BoardRepository(session)
        tasks_repo = TaskRepository(session)
        service = BoardService(board_repo, tasks_repo)

        try:
            board = service.delete_board(id)
            session.commit()
        except BoardNotFoundError:
            return console.print(MessageRenderer.error('Board not found.'))

        console.print(
            MessageRenderer.success(f'Deleted board "{board.name}".'))


@app.command()
def all():
    """Display all boards in a single table.
    """
    with Session(engine) as session:
        board_repo = BoardRepository(session)
        tasks_repo = TaskRepository(session)
        service = BoardService(board_repo, tasks_repo)

        boards = service.list_boards()

        console.clear()
        console.print(BoardRenderer.to_kanban_swimlanes(boards))


@app.command()
def show(id: Annotated[int, typer.Argument(help='Board ID.')]):
    """Display board and its tasks.
    """
    with Session(engine) as session:
        board_repo = BoardRepository(session)
        tasks_repo = TaskRepository(session)
        service = BoardService(board_repo, tasks_repo)

        try:
            board = service.get_board(id)
        except BoardNotFoundError:
            return console.print(MessageRenderer.error('Board not found.'))

        console.clear()
        console.print(BoardRenderer.to_kanban(board))


@app.command()
def clean(id: Annotated[int, typer.Argument(help='Board ID.')],
          force: Annotated[bool, typer.Option(
              '--force', '-f',
              prompt='Are you sure you want to delete completed tasks?',
              help='Force deletion without confirmation.', )] = False):
    """Delete completed tasks from a board.

    If --force is not used, will ask for confirmation.
    """
    if not force:
        return

    with Session(engine) as session:
        board_repo = BoardRepository(session)
        tasks_repo = TaskRepository(session)
        service = BoardService(board_repo, tasks_repo)

        try:
            board = service.clean_completed_tasks(id)
            session.commit()
        except BoardNotFoundError:
            return console.print(MessageRenderer.error('Board not found.'))

        console.clear()
        console.print(BoardRenderer.to_kanban(board))



"""Global app settings and constants.
"""

import os
from pathlib import Path

from .enums import Status


DB_PATH = Path(os.environ.get('KBOARD_HOME', Path.home())) / '.kboard.db'
"""Location to the SQLite file.
"""

STATUS_NAMES: dict[Status, str] = {
    Status.TO_DO: 'To do',
    Status.IN_PROGRESS: 'In progress',
    Status.REVIEW: 'Review',
    Status.COMPLETED: 'Completed',
}
"""Display name for each status."""

STATUS_COLOURS: dict[Status, str] = {
    Status.TO_DO: 'white',
    Status.IN_PROGRESS: 'blue',
    Status.REVIEW: 'magenta',
    Status.COMPLETED: 'green',
}
"""Display colour for each status."""



"""This module exports the service class for the Board model.
"""

from collections.abc import Sequence

from ..exceptions import BoardNotFoundError
from ..models import Board
from ..repos.board_repo import BoardRepository
from ..repos.task_repo import TaskRepository


class BoardService:
    """Application service responsible for board-related use cases.

    This class implements business operations involving Board entities.
    """

    def __init__(self, board_repo: BoardRepository, task_repo: TaskRepository):
        """Initialise the service with repositories.

        :param board_repo: board repository
        :param task_repo: task repository
        """
        self.board_repo = board_repo
        self.task_repo = task_repo

    def list_boards(self) -> Sequence[Board]:
        """Return a list of existing boards.

        :return: list of boards.
        """
        return self.board_repo.list_all()

    def get_board(self, board_id: int) -> Board:
        """Get a Board object by ID or fail if it does not exist.

        :param board_id: board ID to search
        :raises BoardNotFoundError: if the ID does not exist
        :return: board object
        """
        board = self.board_repo.get(board_id)

        if not board:
            raise BoardNotFoundError

        return board

    def create_board(self, name: str) -> Board:
        """Create a new Board object in the database.

        :param name: board name
        :return: board object
        """
        board = Board(name=name)
        self.board_repo.add(board)

        return board

    def rename_board(self, board_id: int, name: str) -> Board:
        """Update the name of a board.

        :param board_id: board ID to search
        :param name: new board name
        :return: board object
        """
        board = self.get_board(board_id)
        board.name = name

        return board

    def delete_board(self, board_id: int) -> Board:
        """Delete a board from the database.

        :param board_id: board ID to search
        :return: board object
        """
        board = self.get_board(board_id)
        self.board_repo.delete(board)

        return board

    def clean_completed_tasks(self, board_id: int) -> Board:
        """Remove all completed tasks from a board.

        :param board_id: board ID to search
        :return: board object
        """
        board = self.get_board(board_id)
        self.task_repo.delete_completed_from_board(board_id)

        return board



"""This module defines the repository class for the Board model.
"""

from collections.abc import Sequence

from sqlalchemy import select
from sqlalchemy.orm import Session

from ..models import Board


class BoardRepository:
    """Repository responsible for persistence operations related to Board
    entities.
    """

    def __init__(self, session: Session):
        """Initialise the repository with a database session.

        :param session: SQLAlchemy session
        """
        self.session = session

    def list_all(self) -> Sequence[Board]:
        """Return a list of all the boards in the database.

        :return: list of boards
        """
        return self.session.execute(select(Board)).scalars().all()

    def get(self, board_id: int) -> Board | None:
        """Retrieve a board object by its ID if it exists.

        :param board_id: id to search
        :return: board object or None
        """
        return self.session.get(Board, board_id)

    def add(self, board: Board) -> None:
        """Add a new board to the session.

        :param board: board object
        """
        self.session.add(board)

    def delete(self, board: Board) -> None:
        """Delete a board from the session.

        :param board: board to delete
        """
        self.session.delete(board)



"""Commands related to the Kanban tasks backlog.
"""

from sqlalchemy.orm import Session
import typer

from ..console import console
from ..db.engine import engine
from ..renderers.board_renderer import BoardRenderer
from ..repos.board_repo import BoardRepository
from ..repos.task_repo import TaskRepository
from ..services.task_service import TaskService


app = typer.Typer()


@app.command()
def backlog():
    """Display tasks from backlog.

    The backlog is composed of tasks with no assigned board.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        board_repo = BoardRepository(session)
        service = TaskService(task_repo, board_repo)

        tasks = service.get_backlog()

        console.clear()
        console.print(BoardRenderer.kanban_from_tasks('Backlog', list(tasks)))



"""Commands related to app configuration.
"""

import typer

from ..console import console
from ..db.init import init_db
from ..renderers.message_renderer import MessageRenderer


app = typer.Typer()


@app.command()
def configure():
    """Create and initialise data file.
    """
    init_db()

    console.print(MessageRenderer.success('Data file created successfully.'))



"""This module defines the enums to be used by database models or CLI
interfaces.
"""

from enum import Enum


class Priority(int, Enum):
    """Task priority.
    """

    LOW = 1
    NORMAL = 2
    HIGH = 3


class Status(int, Enum):
    """Task status.
    """

    TO_DO = 1
    IN_PROGRESS = 2
    REVIEW = 3
    COMPLETED = 4



"""Database-related utility functions.
"""

from .engine import engine
from ..models import Base


def init_db() -> None:
    """Create the database file and create all the tables.
    """
    Base.metadata.create_all(engine)



"""This module exports the renderer class for task objects.
"""

from datetime import date

from rich.panel import Panel

from ..config import STATUS_COLOURS
from ..enums import Priority, Status
from ..models import Task


class TaskRenderer:
    """Class responsible for defining how a task should be displayed.
    """

    @staticmethod
    def _build_subtitle(task: Task) -> str | None:
        """Helper function to generate the subtitle of a task panel.

        :param task: task object
        :return: subtitle as string
        """
        if not task.due_date:
            return None

        today = date.today()

        if task.status == Status.COMPLETED or task.due_date > today:
            colour = 'default'
        elif task.due_date == today:
            colour = 'yellow'
        else:
            colour = 'red'

        return f'[{colour}]{task.due_date}[/]'

    @classmethod
    def to_panel(cls, task: Task) -> Panel:
        """Generate a rich displayable panel representing a Kanban task.

        :param task: task object
        :return: rich panel
        """
        content = task.title

        if task.priority == Priority.LOW:
            content = f'[bright_black]{content}[/]'
        elif task.priority == Priority.HIGH:
            content = f'[yellow]\\[!][/] {content}'

        if task.tag:
            content += f' ([cyan]{task.tag}[/])'

        subtitle = cls._build_subtitle(task)

        return Panel(content, title=str(task.id), title_align='left',
                     border_style=STATUS_COLOURS[task.status],
                     subtitle=subtitle, subtitle_align='right')



"""This module exports the service class for the Task model.
"""

from datetime import datetime

from ..exceptions import BoardNotFoundError, TaskNotFoundError
from ..enums import Priority, Status
from ..models import Task
from ..repos.board_repo import BoardRepository
from ..repos.task_repo import TaskRepository


class TaskService:
    """Application service responsible for task-related use cases.

    This class implements business operations involving Task entities.
    """

    def __init__(self, task_repo: TaskRepository, board_repo: BoardRepository):
        """Initialise the service with repositories.

        :param task_repo: task repository
        :param board_repo: board repository
        """
        self.task_repo = task_repo
        self.board_repo = board_repo

    def get_task(self, task_id: int) -> Task:
        """Get a Task object by ID or fail if it does not exist.

        :param task_id: task ID to search
        :raises TaskNotFoundError: if the ID does not exist
        :return: task object
        """
        task = self.task_repo.get(task_id)

        if not task:
            raise TaskNotFoundError

        return task

    def add_task(self, title: str, priority: Priority, tag: str | None,
                 due_date: datetime | None, board_id: int | None) -> Task:
        """Create a new task in the database.

        :param title: task title
        :param priority: task priority value
        :param tag: task tag
        :param due_date: task due date
        :param board_id: assigned board ID
        :raises BoardNotFoundError: if the board ID does not exist
        :return: task object
        """
        board = None

        if board_id is not None:
            board = self.board_repo.get(board_id)
            if not board:
                raise BoardNotFoundError

        task = Task(title=title, priority=priority, tag=tag, due_date=due_date,
                    board=board)

        self.task_repo.add(task)

        return task

    def edit_task(self, task_id: int, title: str | None,
                  priority: Priority | None, tag: str | None,
                  due_date: datetime | None, board_id: int | None) -> Task:
        """Edit task attributes in the database.

        :param task_id: task ID to search
        :param title: new title
        :param priority: new priority
        :param tag: new tag
        :param due_date: new due date
        :param board_id: new board ID, None to omit, -1 to unassign
        :raises TaskNotFoundError: if the task ID does not exist
        :raises BoardNotFoundError: if the board ID does not exist
        :return: task object
        """
        task = self.get_task(task_id)

        task.update(title=title, priority=priority, tag=tag, due_date=due_date)

        if board_id is not None:
            if board_id == -1:
                task.board = None
            else:
                board = self.board_repo.get(board_id)
                if not board:
                    raise BoardNotFoundError
                task.board = board

        return task

    def move_task(self, task_id: int, steps: int):
        """Update a task status by a number of steps.

        :param task_id: task ID to search
        :param steps: number of steps
        :raises TaskNotFoundError: if the task ID does not exist
        :raises ValueError: if the amount of steps result in an invalid status
        :return: task object
        """
        task = self.get_task(task_id)

        try:
            task.status = Status(task.status + steps)
        except ValueError:
            raise ValueError('Invalid status movement')

        return task

    def delete_task(self, task_id: int):
        """Remove a task from the database.

        :param task_id: task ID to search
        :raises TaskNotFoundError: if the task ID does not exist
        :return: task object
        """
        task = self.get_task(task_id)

        self.task_repo.delete(task)

        return task

    def get_backlog(self):
        """Return a list of unassigned tasks.

        :return: list of tasks.
        """
        return self.task_repo.list_backlog()



"""This module exports a single instance to the database engine.
"""

from sqlalchemy import create_engine

from ..config import DB_PATH


engine = create_engine(f'sqlite:///{DB_PATH}')
"""Database engine."""



"""Commands responsible for managing tasks.
"""

from datetime import datetime
from typing import Annotated

import typer
from sqlalchemy.orm import Session

from ..console import console
from ..db.engine import engine
from ..exceptions import BoardNotFoundError, TaskNotFoundError
from ..models import Priority
from ..renderers.message_renderer import MessageRenderer
from ..renderers.board_renderer import BoardRenderer
from ..repos.board_repo import BoardRepository
from ..repos.task_repo import TaskRepository
from ..services.task_service import TaskService


app = typer.Typer(name='task', help='Manage tasks.', no_args_is_help=True)


@app.command()
def add(title: Annotated[str, typer.Argument(help='Task title.')],
        priority: Annotated[Priority, typer.Option(
            '--priority', '-p', help='Task priority.')] = Priority.NORMAL,
        tag: Annotated[str | None, typer.Option(
            '--tag', '-t', help='Task custom tag.',)] = None,
        due_date: Annotated[datetime | None, typer.Option(
            '--due', '-d', help='Task due date.', formats=['%Y-%m-%d'])] = None,
        board_id: Annotated[int | None, typer.Option(
            '--board', '-b', help='Board ID to assign to the task.')] = None):
    """Add a new task.

    The task can be preassigned to a board using the --board option.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        board_repo = BoardRepository(session)
        service = TaskService(task_repo, board_repo)

        try:
            task = service.add_task(title, priority, tag, due_date, board_id)
        except BoardNotFoundError:
            return console.print(MessageRenderer.error('Board not found.'))

        session.add(task)
        session.commit()

        console.clear()
        if task.board:
            console.print(BoardRenderer.to_kanban(task.board))
        else:
            backlog = service.get_backlog()
            console.print(BoardRenderer.kanban_from_tasks('Backlog', backlog))


@app.command()
def edit(id: Annotated[int, typer.Argument(help='Task ID.')],
         title: Annotated[str | None, typer.Option(
             '--title', help='New task title.')] = None,
         priority: Annotated[Priority | None, typer.Option(
             '--priority', '-p', help='New task priority.')] = None,
         tag: Annotated[str | None, typer.Option(
             '--tag', '-t', help='Task custom tag.')] = None,
         due_date: Annotated[datetime | None, typer.Option(
             '--due', '-d', help='Task due date.', formats=['%Y-%m-%d'])] = None,
         board_id: Annotated[int | None, typer.Option(
        '-b', '--board', help='New board ID (use -1 to unasign).')] = None):
    """Edit existing task attributes.

    All parameters and options from the `add` command are optional here.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        board_repo = BoardRepository(session)
        service = TaskService(task_repo, board_repo)

        try:
            task = service.edit_task(id, title, priority, tag, due_date,
                                     board_id)
            session.commit()
        except TaskNotFoundError:
            return console.print(MessageRenderer.error('Task not found.'))
        except BoardNotFoundError:
            return console.print(MessageRenderer.error('Board not found.'))

        console.clear()
        if task.board:
            console.print(BoardRenderer.to_kanban(task.board))
        else:
            backlog = service.get_backlog()
            console.print(BoardRenderer.kanban_from_tasks('Backlog',
                                                          list(backlog)))


@app.command()
def mv(id: Annotated[int, typer.Argument(help='Task ID.')],
       steps: Annotated[int, typer.Option(
           '--steps', '-s', help='Number of steps to move.')] = 1):
    """Move a task from its current status.

    To customise the direction or number of steps, use the --steps option.

    To move a task backwards the steps must be negative.
    """
    with Session(engine) as session:
        task_repo = TaskRepository(session)
        board_repo = BoardRepository(session)
        service = TaskService(task_repo, board_repo)

        try:
            task = service.move_task(id, steps)
            session.commit()
        except TaskNotFoundError:
            return console.print(MessageRenderer.error('Task not found.'))
        except ValueError:
            return console.print(
                MessageRenderer.error(f'Unable to move {steps} step(s).'))

        console.clear()
        if task.board:
            console.print(BoardRenderer.to_kanban(task.board))
        else:
            backlog = service.get_backlog()
            console.print(BoardRenderer.kanban_from_tasks('Backlog',
                                                          list(backlog)))


@app.command()
def rm(id: Annotated[int, typer.Argument(help='Task ID.')],
       force: Annotated[bool, typer.Option(
           '--force', '-f',
           prompt='Are you sure you want to delete the task?',
           help='Force deletion without confirmation.')] = False):
    """Delete existing task.

    If --force is not used, will ask for confirmation.
    """
    if not force:
        return

    with Session(engine) as session:
        task_repo = TaskRepository(session)
        board_repo = BoardRepository(session)
        service = TaskService(task_repo, board_repo)

        try:
            task = service.delete_task(id)
            board = task.board
            session.commit()
        except TaskNotFoundError:
            return console.print(MessageRenderer.error('Task not found.'))

        console.clear()
        if board:
            console.print(BoardRenderer.to_kanban(board))
        else:
            backlog = service.get_backlog()
            console.print(BoardRenderer.kanban_from_tasks('Backlog',
                                                          list(backlog)))



"""This module defines the repository class for the Task model.
"""

from collections.abc import Sequence

from sqlalchemy import delete, select
from sqlalchemy.orm import Session

from ..enums import Status
from ..models import Task


class TaskRepository:
    """Repository responsible for persistence operations related to Task
    entities.
    """

    def __init__(self, session: Session):
        """Initialise the repository with a database session.

        :param session: SQLAlchemy session
        """
        self.session = session

    def get(self, task_id: int) -> Task | None:
        """Retrieve a task object by its ID if it exists.

        :param task_id: id to search
        :return: task object or None
        """
        return self.session.get(Task, task_id)

    def add(self, task: Task) -> None:
        """Add a new task to the session.

        :param task: task object
        """
        self.session.add(task)

    def delete(self, task: Task) -> None:
        """Delete a task from the session.

        :param task: task to delete
        """
        self.session.delete(task)

    def list_backlog(self) -> Sequence[Task]:
        """Return a list of unassigned tasks.

        :return: list of tasks
        """
        return self.session.execute(
            select(Task).where(Task.board_id.is_(None))
        ).scalars().all()

    def delete_completed_from_board(self, board_id: int) -> None:
        """Delete completed tasks from the database.

        :param board_id: board ID to clear
        """
        self.session.execute(
            delete(Task).where(Task.board_id == board_id,
                               Task.status == Status.COMPLETED)
        )



"""Application CLI entry point.
"""

import typer

from .commands import backlog, board, configure, task


app = typer.Typer(no_args_is_help=True,
                  context_settings={'help_option_names': ['-h', '--help']})

app.add_typer(configure.app)
app.add_typer(backlog.app)
app.add_typer(board.app)
app.add_typer(task.app)


if __name__ == '__main__':
    app()



"""This module exports the renderer class for status messages.
"""

from rich.panel import Panel


class MessageRenderer:
    """Renderer class to generate rich printable status messages.
    """

    @staticmethod
    def success(message: str) -> Panel:
        """Generate a success message.

        :param message: panel content
        :return: panel object
        """
        return Panel(message, title='Success', title_align='left',
                     border_style='green')

    @staticmethod
    def error(message: str) -> Panel:
        """Generate an error message.

        :param message: panel content
        :return: panel object
        """
        return Panel(message, title='Error', title_align='left',
                     border_style='red')



"""This module declares all the database models and their relationships.
"""

from datetime import date
from typing import overload
from sqlalchemy import ForeignKey
from sqlalchemy.orm import DeclarativeBase, Mapped, mapped_column, relationship

from .enums import Priority, Status


class Base(DeclarativeBase):
    ...


class Board(Base):
    """Container for multiple tasks that represents a Kanban board.
    """

    __tablename__ = 'boards'

    id: Mapped[int] = mapped_column(primary_key=True)
    name: Mapped[str]

    tasks: Mapped[list['Task']] = relationship(back_populates='board',
                                               cascade='all, delete',
                                               passive_deletes=True)


class Task(Base):
    """Unit of work that can be moved across a board.
    """

    __tablename__ = 'tasks'

    id: Mapped[int] = mapped_column(primary_key=True)
    title: Mapped[str]
    priority: Mapped[Priority]
    tag: Mapped[str] = mapped_column(default='')
    status: Mapped[Status] = mapped_column(default=Status.TO_DO)
    due_date: Mapped[date | None]
    board_id: Mapped[int | None] = mapped_column(
        ForeignKey('boards.id', ondelete='CASCADE'))

    board: Mapped[Board | None] = relationship(back_populates='tasks')

    @overload
    def update(self, *, title: str | None = None,
               priority: Priority | None = None, tag: str | None = None,
               status: Status | None = None, due_date: date | None) -> None:
        """Update multiple instance attributes in a single call.

        To skip a field leave is as ``None``.

        :param title: new title
        :param priority: new priority
        :param tag: new tag
        :param status: new status
        """
        ...

    @overload
    def update(self, **kwargs) -> None:
        ...

    def update(self, **kwargs) -> None:
        for attr, value in kwargs.items():
            if value is not None:
                setattr(self, attr, value)
